import { exec } from 'child_process';
import debug from 'debug';
import fs from 'fs';
import { join } from 'path';
import PgBoss from 'pg-boss';
import enqueue from './enqueue.js';
import withDatabaseClient from './withDatabaseClient.js';

const log = debug('catirpel:run');

const listCaterpillars = async caterpillarRoot => {
  const files = await fs.promises.readdir(caterpillarRoot);
  const caterpillars = [...new Set(files.map(x => x.split('.')[0]))];
  return caterpillars.map(caterpillar => ({
    name: caterpillar,
    path: join(caterpillarRoot, `${caterpillar}.js`),
    seedPath: files
      .filter(x => x === `${caterpillar}.seed.js`)
      .map(x => join(caterpillarRoot, x))[0],
  }));
};

const runModule = async path => {
  const contents = await fs.promises.readFile(path, 'utf8');
  return eval(contents);
};

const seed = async (boss, queue, seedPath) => {
  log('seeding', queue, 'via', seedPath);
  const messages = await runModule(seedPath);
  const jobIds = (
    await Promise.all(messages.map(message => enqueue(boss, queue, message)))
  ).filter(Boolean);
  log('enqueued', jobIds.length, 'jobs', jobIds);
  return queue;
};

const process = (caterpillarPath, data) => {
  return new Promise((resolve, reject) => {
    exec(
      `node ${caterpillarPath} '${JSON.stringify(data)}'`,
      {
        env: {},
        shell: true,
      },
      (error, stdout) => {
        if (error) {
          new Error('Processor error');
          reject(error);
        } else {
          resolve(stdout);
        }
      }
    );
  });
};

const run = async caterpillarRoot => {
  // Initialise caterpillars
  const caterpillars = await listCaterpillars(caterpillarRoot);
  const caterpillarMap = caterpillars.reduce((acc, x) => {
    acc[x.name] = x;
    return acc;
  }, {});
  log('found', caterpillars.length, 'caterpillars', caterpillars);

  // Initialise queue
  const boss = new PgBoss({});
  boss.on('error', error => {
    log('queue error', error);
  });
  boss.on('monitor-states', log);
  await boss.start();
  await boss.onComplete('*', job => {
    if (job.data.state === 'failed') {
      log('abandoned message', job);
    }
  });
  log('initialised queue');

  await boss.deleteAllQueues();

  // Create data tables
  await withDatabaseClient(client =>
    Promise.all(
      caterpillars.map(x =>
        client.query(
          `create table if not exists public.${x.name} (
            id text not null primary key,
            updated_at timestamp with time zone default now() not null,
            domain text,
            document jsonb not null
          )`
        )
      )
    )
  );
  log('created data tables');

  // Retrieve queue stats
  const stats = await Promise.all(
    caterpillars.map(async ({ name }) => ({
      caterpillar: name,
      queueSize: await boss.getQueueSize(name),
    }))
  );
  log('retrieved queue statistics', stats);

  // Seed
  await Promise.all(
    stats
      .filter(x => x.queueSize === 0)
      .map(x => caterpillarMap[x.caterpillar])
      .filter(x => x.seedPath)
      .map(x => seed(boss, x.name, x.seedPath))
  );

  // Crawl
  await boss.subscribe('catirpel:*', async job => {
    const caterpillar = caterpillarMap[job.name.replace('catirpel:', '')];
    log(caterpillar.name, 'received a message', job.data);
    try {
      const result = JSON.parse(await process(caterpillar.path, job.data));
      log(
        'got',
        result.messages.length,
        'messages and',
        result.documents.length,
        'documents'
      );

      // Enqueue messages
      await Promise.all(
        result.messages.map(message => enqueue(boss, caterpillar.name, message))
      );

      // Insert/update documents
      await withDatabaseClient(client =>
        Promise.all(
          result.documents.map(({ domain, id, ...data }) =>
            client.query(
              `insert into ${caterpillar.name} values($1, now(), $2, $3)
              on conflict (id) do update set
                updated_at = now(),
                domain = EXCLUDED.domain,
                document = EXCLUDED.document`,
              [id, domain, data]
            )
          )
        )
      );

      job.done();
    } catch (error) {
      log(caterpillar.name, 'fell into a hole', error);
      throw error;
    }
  });
};

export default run;
